from .species import Species
from ..quantities import *


class C5H10(Species):
    def __init__(self, T=None):
        self.mol_weight = 70.135
        self.c_atoms = 5
        self.h_atoms = 10
        super().__init__(T)



import os, requests, pandas as pd

from catrxneng import conf, utils
from catrxneng.utils import Influx, Time
from catrxneng.quantities import *


class Expt:

    @property
    def time_on_stream_hr(self):
        try:
            return (self.time_series_data["timestamp"] - self.t0) / 3600
        except AttributeError:
            raise AttributeError("Experiment has no t0 value assigned.")

    @property
    def steady_state_steps(self):
        return [step for step in self.steps if step.step_name == "steadyState"]

    def __init__(
        self,
        expt_type,
        reactor_class,
        kinetic_model,
        step_class,
        catalyst,
        mcat,
        unit,
        lab_notebook_id,
        T=None,
        p0=None,
        whsv=None,
    ):
        self.expt_type=expt_type
        self.reactor_class = reactor_class
        self.kinetic_model = kinetic_model
        self.step_class = step_class
        self.catalyst = catalyst
        self.mcat = mcat
        self.unit = unit
        self.lab_notebook_id = lab_notebook_id
        self.T = T
        self.p0 = p0
        self.whsv = whsv
        self.steps = []

    def add_step(self, step_name, start=None, end=None, T=None, p0=None, whsv=None):
        if start is None:
            start = self.steps[-1].end
        if end is None:
            end = start + TimeDelta(hr=1)
        if T is None:
            T = self.T
        if p0 is None:
            p0 = self.p0
        if whsv is None:
            whsv = self.whsv
        if T is None or p0 is None or whsv is None:
            raise ValueError("Step has a missing condition.")
        step = self.step_class(step_name, len(self.steps) + 1, start, end)
        step.attach_reactor(
            self.reactor_class, self.kinetic_model, T, p0, whsv, self.mcat
        )
        self.steps.append(step)

    def simulate(self, dt_sec, std_dev=None):
        for step in self.steps:
            step.simulate(dt_sec, std_dev)
        dataframes = [step.time_series_data for step in self.steps]
        self.time_series_data = pd.concat(dataframes, ignore_index=True)
        self._compute_tos()

    def upload_data(self, bucket, measurement):
        influx = Influx(
            url=os.getenv("INFLUXDB_URL"),
            org=os.getenv("INFLUXDB_ORG"),
            bucket=bucket,
            measurement=measurement,
        )
        # conf = utils.getconf(self.unit, "tags")
        # tags = [
        #     conf[data_id]
        #     for data_id in self.time_series_data.columns
        #     if data_id in conf
        # ]
        df = self.time_series_data.rename(columns=utils.getconf(self.unit, "tags"))
        influx.upload_dataframe(dataframe=df)

    def _compute_tos(self):
        for step in self.steps:
            if step.step_name == "steadyState":
                start = step.start.UET
                break
        self.tos = self.time_series_data["timestamp"] - start

    def upload_to_emp(self, host, dt_sec, notes=None):
        # url = "http://192.168.215.193:5001/10057/api"
        endpoint = "/api/create_simulated_expt/10057"
        url = host + endpoint
        params = {
            "expt_type": self.expt_type,
            "lab_notebook_id": self.lab_notebook_id,
            "unit": self.unit,
            "material__common_name": self.catalyst,
            "sample_mass": self.mcat.g,
            "start__ET_str": self.steps[0].start.ET_str,
            "end__ET_str": self.steps[-1].end.ET_str,
            "dt_sec": dt_sec,
            "notes": notes,
        }
        return requests.post(url, json=params, timeout=10).json()



from catrxneng import utils
from .species import Species
from ..quantities import *


class CO(Species):
    def __init__(self, T=None):
        self.mol_weight = 28
        self.Hf_298_gas = Energy(kJmol=-110.53)
        self.S_298_gas = Entropy(JmolK=197.66)
        self.nist_thermo_params = [
            {
                "min_temp_K": 298,
                "max_temp_K": 1300,
                "phase": "gas",
                "A": 25.56759,
                "B": 6.096130,
                "C": 4.054656,
                "D": -2.671301,
                "E": 0.131021,
                "F": -118.0089,
                "G": 227.3665,
                "H": -110.5271,
            }
        ]
        super().__init__(T)

    def Hf_gas(self, T=None):
        if not T:
            T = self.T
        return utils.Hf_shomate(T, self.nist_thermo_params[0])

    def S_gas(self, T=None):
        if not T:
            T = self.T
        return utils.S_shomate(T, self.nist_thermo_params[0])

    def Cp_gas(self, T=None):
        if not T:
            T = self.T
        return utils.Cp_shomate(T, self.nist_thermo_params[0])


from .species import Species
from ..quantities import *
from catrxneng import utils


class H2(Species):
    def __init__(self, T=None):
        self.mol_weight = 2
        self.Hf_298_gas = Energy(kJmol=0)
        self.S_298_gas = Entropy(JmolK=130.68)
        self.nist_thermo_params = [
            {
                "min_temp_K": 298,
                "max_temp_K": 1000,
                "phase": "gas",
                "A": 33.066178,
                "B": -11.363417,
                "C": 11.432816,
                "D": -2.772874,
                "E": -0.158558,
                "F": -9.980797,
                "G": 172.707974,
                "H": 0,
            }
        ]
        super().__init__(T)

    def Hf_gas(self, T=None):
        if not T:
            T = self.T
        return utils.Hf_shomate(T, self.nist_thermo_params[0])

    def S_gas(self, T=None):
        if not T:
            T = self.T
        return utils.S_shomate(T, self.nist_thermo_params[0])

    def Cp_gas(self, T=None):
        if not T:
            T = self.T
        return utils.Cp_shomate(T, self.nist_thermo_params[0])



import numpy as np
from scipy.optimize import least_squares

from catrxneng.quantities import *


class KineticModel:

    def __init__(self, limiting_reactant, T):
        self.limiting_reactant = limiting_reactant
        self.fugacity_coeff = np.ones(len(self.components))
        self.std_state_fugacity = std_state_fugacity.bar
        self.T = T
        if self.T is not None:
            self.compute_temp_dependent_constants()

    def compute_temp_dependent_constants(self):
        pass

    def _compute_final_molfrac(self, initial_moles, extent):
        delta_moles = self.stoich_coeff.T @ extent
        final_moles = initial_moles + delta_moles
        total_final_moles = np.sum(final_moles)
        molfrac = final_moles / total_final_moles
        return molfrac, delta_moles

    def _compute_extent_bounds(self, initial_moles):
        lower_bounds = [0] * len(self.reactions)
        upper_bounds = []
        for i in range(len(self.reactions)):
            stoich_lim = self.stoich_coeff[i][self.comp_idx[self.limiting_reactant]]
            max_extent = initial_moles[self.comp_idx[self.limiting_reactant]] / abs(
                stoich_lim
            )
            upper_bounds.append(max_extent)
        return (lower_bounds, upper_bounds)

    def _equilibrium_objective(self, extent, P, initial_moles, Keq):
        molfrac = self._compute_final_molfrac(initial_moles, extent)[0]
        molfrac[molfrac < 0] = 0.00001
        p = molfrac * P
        fugacity = self.fugacity_coeff * p
        activity = fugacity / self.std_state_fugacity
        K_calc = np.prod(activity**self.stoich_coeff, axis=1)
        return (np.log(K_calc) - np.log(Keq)) ** 2

    def equilibrate(self, p0, T, initial_guesses=None, allow_component_mismatch=False):
        if T and T != self.T:
            self.T = T
            self.compute_temp_dependent_constants()
        initial_total_moles = 100
        if allow_component_mismatch:
            p0_bar = [p0[comp].bar for comp in self.components]
            p0 = Pressure(bar=p0_bar, keys=self.components)
        P = np.sum(p0)
        initial_molfrac = p0 / P
        initial_moles = (initial_molfrac * initial_total_moles).si
        if not initial_guesses:
            num_rxns = len(self.reactions)
            initial_guess = (
                initial_moles[self.comp_idx[self.limiting_reactant]] / num_rxns
            )
            initial_guesses = np.ones(num_rxns) * initial_guess / 2

        def objective(extent):
            return self._equilibrium_objective(extent, P.bar, initial_moles, self.Keq)

        solution = least_squares(
            objective,
            initial_guesses,
            bounds=self._compute_extent_bounds(initial_moles),
            method="trf",
            ftol=1e-10,
            max_nfev=1000,
        )
        extent = solution.x
        eq_molfrac, delta_moles = self._compute_final_molfrac(initial_moles, extent)
        delta = delta_moles[self.comp_idx[self.limiting_reactant]]
        initial = initial_moles[self.comp_idx[self.limiting_reactant]]
        self.eq_conversion = Unitless(si=-delta / initial)
        self.eq_molfrac = Unitless(si=eq_molfrac, keys=self.components)

    def carbon_balance(self, reactor):
        carbon_in = np.sum(reactor.F0 * self.carbon_atoms)
        carbon_out = np.sum(reactor.F * self.carbon_atoms[:, np.newaxis], axis=0)
        return carbon_out / carbon_in



import pandas as pd, numpy as np

from .step import Step
from catrxneng import utils
from catrxneng.quantities import *


class TandemStep(Step):

    def __init__(self, step_name, step_num, start, end):
        super().__init__(step_name, step_num, start, end)

    def attach_reactor(self, reactor_class, kinetic_model, T, p0, whsv, mcat):
        self.reactor = reactor_class(
            kinetic_model=kinetic_model,
            p0=p0,
            whsv=whsv,
            T=T,
            limiting_reactant="co2",
            mcat=mcat,
        )

    def simulate(self, dt_sec, std_dev=None):
        if std_dev is None:
            std_dev = {"temp": 0.3, "pressure": 0.1, "mfc": 0.5, "gc": 0.005}
        df = pd.DataFrame()
        df["timestamp"] = np.arange(self.start.UET, self.end.UET + dt_sec, dt_sec)
        num_points = df["timestamp"].size
        df["step_name"] = self.step_name
        df["step_num"] = self.step_num
        self.reactor.solve()
        df["bed_temp"] = self.reactor.T.C + np.random.normal(
            loc=0, scale=std_dev["temp"], size=num_points
        )
        df["pressure"] = self.reactor.P.bar + np.random.normal(
            loc=0, scale=std_dev["pressure"], size=num_points
        )
        df["n2_mfc"] = self.reactor.F0["inert"].smLmin + np.random.normal(
            loc=0, scale=std_dev["mfc"], size=num_points
        )
        df["co2_mfc"] = self.reactor.F0["co2"].smLmin + np.random.normal(
            loc=0, scale=std_dev["mfc"], size=num_points
        )
        df["h2_mfc"] = self.reactor.F0["h2"].smLmin + np.random.normal(
            loc=0, scale=std_dev["mfc"], size=num_points
        )

        for component in self.reactor.y.keys:
            df[f"{component}_gc_conc"] = self.reactor.y[-1][
                component
            ].pct * np.random.normal(loc=1, scale=std_dev["gc"], size=num_points)
            if component == "inert":
                df = df.rename(columns={"inert_gc_conc": "n2_gc_conc"})

        self.time_series_data = df



import os, numpy as np, pandas as pd
from importlib import import_module

from .influx import Influx
from .time import Time


def get_keys(obj1, obj2):
    if getattr(obj1, "keys", None) == getattr(obj2, "keys", None):
        keys = getattr(obj1, "keys", None)
    if getattr(obj1, "keys", None) is not None:
        keys = obj1.keys
    if getattr(obj2, "keys", None) is not None:
        keys = obj2.keys
    try:
        if keys is None:
            return None
        else:
            return keys.copy()
    except NameError:
        raise ValueError("Quantities have mismatching keys.")


def rate_const(T, Ea, k0=None, kref=None, Tref=None):
    from ..quantities import R, RateConstant

    if kref and Tref:
        k0 = kref.si / (np.exp(-Ea / (R * Tref))).si
        k0 = RateConstant(si=k0, order=kref.order)

    k = k0.si * np.exp(-Ea / (R * T)).si
    return RateConstant(si=k, order=k0.order)


# def vant_hoff_eq_const(T, dH, Kref, Tref):
#     from ..quantities import R, EquilibriumConstant

#     Keq = Kref.si * np.exp(dH / R * (1 / Tref - 1 / T)).si
#     return EquilibriumConstant(si=Keq, order=Kref.order)


def to_float(value):
    if isinstance(value, (np.ndarray, pd.Series)):
        return value.astype(float)
    if isinstance(value, int):
        return float(value)
    return value


def plot_info_box(text):
    if text is None:
        return None
    text = text.replace("\\n", "\n")
    text = text.strip("\n")
    text = text.replace("\n", "<br>")
    annotations = [
        dict(
            x=0.5,
            y=-0.35,
            xref="paper",
            yref="paper",
            text=text,
            showarrow=False,
            font=dict(size=12, style="italic", weight="bold"),
        )
    ]
    return annotations


def divide(x, y):
    if isinstance(y, (np.ndarray, pd.Series)):
        y_safe = y.copy()
        y_safe[y_safe == 0] = np.nan
    else:
        y_safe = np.nan if y == 0 else y

    return x / y_safe


def getconf(conf_name, variable):
    conf_module_path = os.getenv("CONF_MODULE_PATH")
    # module_path = f"catrxneng.conf.{conf_name}"
    module_path = f"{conf_module_path}.{conf_name}"
    # try:
    conf_module = import_module(module_path)
    # except ModuleNotFoundError as e:
    #     raise ImportError(f"Configuration module '{module_path}' not found.") from e
    # try:
    return getattr(conf_module, variable)
    # except AttributeError as e:
    #     raise AttributeError(
    #         f"Variable '{variable}' not found in '{module_path}'."
    #     ) from e


def Hf_shomate(T, params):
    from catrxneng.quantities import Energy

    t = T.K / 1000
    dHf = (
        params["A"] * t
        + params["B"] * t**2 / 2
        + params["C"] * t**3 / 3
        + params["D"] * t**4 / 4
        - params["E"] / t
        + params["F"]
        - params["H"]
    )  # kJ/mol
    return Energy(kJmol=dHf)


def S_shomate(T, params):
    from catrxneng.quantities import Entropy

    t = T.K / 1000
    S = (
        params["A"] * np.log(t)
        + params["B"] * t
        + params["C"] * t**2 / 2
        + params["D"] * t**3 / 3
        - params["E"] / (2 * t**2)
        + params["G"]
    )  # J/mol/K
    return Entropy(JmolK=S)


def Cp_shomate(T, params):
    from catrxneng.quantities import HeatCapacity

    t = T.K / 1000
    Cp = (
        params["A"]
        + params["B"] * t
        + params["C"] * t**2
        + params["D"] * t**3
        + params["E"] / (t**2)
    )
    return HeatCapacity(JmolK=Cp)


def vant_hoff_eqn(x_ref, dH, T, Tref):
    from catrxneng.quantities import R

    return x_ref * np.exp(dH / R * (1 / Tref - 1 / T))



from .species import Species
from ..quantities import *

class Graphite(Species):
    def __init__(self, T=None):
        self.mol_weight = 12
        self.S_298 = Entropy(JmolK=5.6)
        super().__init__(T)
    
    def Cp(self, T):
        return HeatCapacity(JmolK=8.23)



import numpy as np

from ..kinetic_model import KineticModel
from ..co2_to_ch3oh import Ghosh2021
from ..mto import Lee2019
from catrxneng.quantities import *


class GhoshLee(KineticModel):

    def __init__(self, co2_mto_catalyst_ratio, limiting_reactant="co2", T=None):
        self.mass_frac_co2_to_meoh_catalyst = co2_mto_catalyst_ratio / (
            1 + co2_mto_catalyst_ratio
        )
        self.mass_frac_mto_catalyst = 1 - self.mass_frac_co2_to_meoh_catalyst
        self.km1 = Ghosh2021(T=T)
        self.km2 = Lee2019(T=T)
        self.components = [
            "co2",
            "h2",
            "ch3oh",
            "h2o",
            "co",
            "ch4",
            "c2h4",
            "c3h6",
            "c3h8",
            "c4h8",
            "c4",
            "c5+",
            "inert",
        ]
        self.carbon_atoms = np.array([1, 0, 1, 0, 1, 1, 2, 3, 3, 4, 4, 5, 0])
        self.km1_map = [self.components.index(comp) for comp in self.km1.components]
        self.km2_map = [self.components.index(comp) for comp in self.km2.components]
        super().__init__(limiting_reactant, T)

    def set_temp(self, T):
        self.T = T
        self.km1.T = T
        self.km2.T = T

    def compute_temp_dependent_constants(self):
        self.km1.compute_temp_dependent_constants()
        self.km2.compute_temp_dependent_constants()

    def rate_equations(self, p_array):
        r1 = self.km1.rate_equations(p_array[self.km1_map])
        r1 = r1 * self.mass_frac_co2_to_meoh_catalyst
        r2 = self.km2.rate_equations(p_array[self.km2_map])
        r2 = r2 * self.mass_frac_mto_catalyst
        return np.array(
            [
                r1[0],  # co2
                r1[1],  # h2
                r1[2] + r2[7],  # meoh
                r1[3] + r2[8],  # h2o
                r1[4],  # co
                r1[5] + r2[0],  # ch4
                r2[1],  # c2h4
                r2[2],  # c3h6
                r2[3],  # c3h8
                r2[4],  # c4h8
                r2[5],  # c4
                r2[6],  # c5+
                0.0,  # inert
            ]
        )



import numpy as np

from ..kinetic_model import KineticModel
from catrxneng.utils import vant_hoff_eqn
from catrxneng.quantities import *


class Ghosh2022MTO(KineticModel):
    def __init__(self, limiting_reactant="ch3oh", T=None, kref=None, Ea=None):
        self.Tref = Temperature(C=320)
        self.components = [
            "ch3oh",
            "h2",
            "h2o",
            "c2h4",
            "c3h6",
            "c4h8",
            "c2h6",
            "c3h8",
            "c4h10",
            "c5-8",
            "c9+",
            "inert",
        ]
        self.f8 = {
            "c2h4": 5e-2,
            "c3h6": 1.0,
            "c4h8": 2.6e-1,
        }
        self.f9 = {
            "c2h4": 5.5e-1,
            "c3h6": 1.0,
            "c4h8": 4.5e-1,
        }
        self.carbon_atoms = np.array([1, 0, 0, 2, 3, 4, 2, 3, 4, 6.5, 10.5, 0])
        self.kref = kref
        if self.kref is None: 
            self.kref = np.array([5.9e-1, 7e-2, 6e-1, 5.9e-2, 8e-2, 8.3e-1, 6e-2, 8.2e-1])
        self.Ea = Ea
        if self.Ea is None:
            self.Ea = np.array([70.0, 16.0, 17.0, 69.0, 109.0, 150.0, 170.0, 25.0])
        self.order = np.array([1, 2, 2, 2, 2, 1, 1, 2])
        self.comp_idx = {comp: i for i, comp in enumerate(self.components)}
        super().__init__(limiting_reactant, T)

    def compute_temp_dependent_constants(self):
        self.K_ch3oh = vant_hoff_eqn(1.3e1, Energy(kJmol=-2e-1), self.T, self.Tref).si
        self.k = np.array(
            [
                RateConstant(
                    molskgcatbar=kref, Ea=Energy(kJmol=Ea), Tref=self.Tref, order=order
                )(self.T).molhgcatbar
                for kref, Ea, order in zip(self.kref, self.Ea, self.order)
            ]
        )

    def rate_equations(self, p_array):
        p_ch3oh = p_array[self.comp_idx["ch3oh"]]
        p_h2 = p_array[self.comp_idx["h2"]]
        p_h2o = p_array[self.comp_idx["h2o"]]
        p_c2h4 = p_array[self.comp_idx["c2h4"]]
        p_c3h6 = p_array[self.comp_idx["c3h6"]]
        p_c4h8 = p_array[self.comp_idx["c4h8"]]
        p_c58 = p_array[self.comp_idx["c5-8"]]
        inhib = 1 + self.K_ch3oh * (p_ch3oh + p_h2o)

        n = 4
        r4 = self.k[4 - n] * p_ch3oh / inhib
        r5 = self.k[5 - n] * p_ch3oh * p_c2h4 / inhib
        r6 = self.k[5 - n] * p_ch3oh * p_c3h6 / inhib
        r7 = self.k[5 - n] * p_ch3oh * p_c4h8 / inhib
        r8_1 = self.k[8 - n] * p_c2h4 * p_h2 / inhib * self.f8["c2h4"]
        r8_2 = self.k[8 - n] * p_c3h6 * p_h2 / inhib * self.f8["c3h6"]
        r8_3 = self.k[8 - n] * p_c4h8 * p_h2 / inhib * self.f8["c4h8"]
        r9_1 = self.k[9 - n] * p_c2h4 / inhib * self.f9["c2h4"]
        r9_2 = self.k[9 - n] * p_c3h6 / inhib * self.f9["c3h6"]
        r9_3 = self.k[9 - n] * p_c4h8 / inhib * self.f9["c4h8"]
        r10_1 = self.k[10 - n] * p_c58 / inhib * self.f9["c2h4"]
        r10_2 = self.k[10 - n] * p_c58 / inhib * self.f9["c3h6"]
        r10_3 = self.k[10 - n] * p_c58 / inhib * self.f9["c4h8"]
        r11_1 = self.k[11 - n] * p_c2h4 * p_c58 / inhib * self.f9["c2h4"]
        r11_2 = self.k[11 - n] * p_c3h6 * p_c58 / inhib * self.f9["c3h6"]
        r11_3 = self.k[11 - n] * p_c4h8 * p_c58 / inhib * self.f9["c4h8"]

        return np.array(
            [
                -9 * r4 - 2 * r5 - 3 * r6 - 4 * r7,  # ch3oh
                -1 * r8_1 - r8_2 - r8_3,  # h2
                9 * r4 + 2 * r5 + 3 * r6 + 4 * r7,  # h2o
                r4 + r5 - r8_1 - r9_1 + r10_1 - r11_1,  # c2h4
                r4 + r6 - r8_2 - r9_2 + r10_2 - r11_2,  # c3h6
                r4 + r7 - r8_3 - r9_3 + r10_3 - r11_3,  # c4h8
                r8_1,  # c2h6
                r8_2,  # c3h8
                r8_3,  # c4h10
                0.307 * r9_1
                + 0.461 * r9_2
                + 0.615 * r9_3
                - 0.307 * r10_1
                - 0.461 * r10_2
                - 0.615 * r10_3
                - r11_1
                - r11_2
                - r11_3,  # c5-8
                0.809 * r11_1 + 0.904 * r11_2 + r11_3,  # c9+
                0.0,  # inert
            ]
        )



from catrxneng import utils
from .species import Species
from ..quantities import *


class CH4(Species):
    def __init__(self, T=None):
        self.mol_weight = 16
        self.Hf_298_gas = Energy(kJmol=-74.6)
        self.S_298_gas = Entropy(JmolK=186.3)
        self.nist_thermo_params = [
            {
                "min_temp_K": 298,
                "max_temp_K": 1300,
                "phase": "gas",
                "A": -0.703029,
                "B": 108.4773,
                "C": -42.52157,
                "D": 5.862788,
                "E": 0.678565,
                "F": -76.84376,
                "G": 158.7163,
                "H": -74.87310,
            }
        ]
        super().__init__(T)

    def Hf_gas(self, T=None):
        if not T:
            T = self.T
        return utils.Hf_shomate(T, self.nist_thermo_params[0])

    def S_gas(self, T=None):
        if not T:
            T = self.T
        return utils.S_shomate(T, self.nist_thermo_params[0])

    def Cp_gas(self, T=None):
        if not T:
            T = self.T
        return utils.Cp_shomate(T, self.nist_thermo_params[0])


import numpy as np
from catrxneng import utils


class GasMixture:

    def __init__(self, y=None, p=None):
        self.update(y, p)

    def update(self, y=None, p=None):
        from catrxneng.species import CLASS_MAP
        from ..quantities import Pressure, Unitless

        self.y = y
        self.p = p
        if y:
            comp_list = list(y.keys)
        if p:
            comp_list = list(p.keys)
            self.P = Pressure(si=np.sum([p.si]))
            self.y = utils.divide(p, self.P)
        active_comp_list = comp_list.copy()
        active_comp_list.remove("inert")
        self.components = {
            comp_id: CLASS_MAP[comp_id]() if comp_id in CLASS_MAP else None
            for comp_id in comp_list
        }
        sum_active = np.sum(
            [
                molfrac
                for comp_id, molfrac in zip(comp_list, self.y.si)
                if comp_id.lower() != "inert"
            ]
        )
        y_active = [self.y[comp_id].si for comp_id in active_comp_list]
        self.y_active = Unitless(
            si=utils.divide(y_active, sum_active), keys=active_comp_list
        )
        self.avg_mol_weight = np.sum(
            [
                self.components[comp_id].mol_weight * self.y_active[comp_id].si
                for comp_id in active_comp_list
                if self.components[comp_id]
            ]
        )



from .species import Species
from ..quantities import *


class C4H8(Species):
    def __init__(self, T=None):
        self.mol_weight = 56.108
        self.c_atoms = 4
        self.h_atoms = 8
        super().__init__(T)



import numpy as np

from ..kinetic_model import KineticModel
from catrxneng.quantities import *
from catrxneng.reactors import PFR


class Lee2019(KineticModel):
    def __init__(self, limiting_reactant="ch3oh", T=None, kref=None, Ea=None):
        self.Tref = Temperature(C=450)
        self.components = [
            "ch4",
            "c2h4",
            "c3h6",
            "c3h8",
            "c4h8",
            "c4",
            "c5+",
            "ch3oh",
            "h2o",
            "inert",
        ]
        self.comp_idx = {comp: i for i, comp in enumerate(self.components)}
        self.carbon_atoms = np.array([1, 2, 3, 3, 4, 4, 5, 1, 0, 0])
        self.kref = kref
        if self.kref is None:
            self.kref = np.array([0.16, 7.16, 7.35, 0.39, 2.26, 0.29, 0.9, 0.53, 0.36])
        self.Ea = Ea
        if self.Ea is None:
            self.Ea = np.array([71.7, 44.15, 14.02, 3.8, 6.24, 6.24, 10.0, 13.02, 4.04])
        self.K_h2o = 6.05
        super().__init__(limiting_reactant, T)

    def compute_temp_dependent_constants(self):
        self.k = np.array(
            [
                RateConstant(
                    Lgcatmin=kref, Ea=Energy(kJmol=Ea), Tref=self.Tref, order=1
                )(self.T).molhgcatbar
                for kref, Ea in zip(self.kref, self.Ea)
            ]
        )

    def rate_equations(self, p_array):
        p_meoh = p_array[self.comp_idx["ch3oh"]]
        p_h2o = p_array[self.comp_idx["h2o"]]
        p_c2h4 = p_array[self.comp_idx["c2h4"]]
        p_c3h6 = p_array[self.comp_idx["c3h6"]]
        P = np.sum(p_array)
        if P == 0:
            y_h2o = 0
        else:
            y_h2o = p_h2o / P
        theta_w = 1 / (1 + self.K_h2o * y_h2o)
        r = np.array(
            [
                self.k[0] * theta_w * p_meoh,
                self.k[1] * theta_w * p_meoh,
                self.k[2] * theta_w * p_meoh,
                self.k[3] * theta_w * p_meoh,
                self.k[4] * theta_w * p_meoh,
                self.k[5] * theta_w * p_meoh,
                self.k[6] * theta_w * p_meoh,
                self.k[7] * theta_w * p_c2h4,
                self.k[8] * theta_w * p_c3h6,
            ]
        )
        return np.array(
            [
                r[0],  # ch4
                (r[1] - r[7]) / 2,  # c2h4
                (r[2] + r[7] - r[8]) / 3,  # c3h6
                r[3] / 3,  # c3h8
                (r[4] + r[8]) / 4,  # c4h8
                r[5] / 4,  # c4
                r[6] / 5,  # c5+
                np.sum([-r[i] for i in range(7)]),  # meoh
                np.sum([r[i] for i in range(7)]),  # h2o
                0.0,  # inert
            ]
        )



import numpy as np

from catrxneng import utils
from . import Quantity, Temperature


class RateConstant(Quantity):
    def __init__(self, Ea, order, Tref=None, **kwargs):
        self.Ea = Ea
        self.order = order
        self.Tref = Tref
        if self.Tref:
            self.T = self.Tref
        else:
            self.T = Temperature(K=298)
        super().__init__(**kwargs)
        if self.Tref:
            self.kref_si = self.si
        else:
            self.k0_si = self.si

    @property
    def molskgcatPa(self):
        return self.si

    @molskgcatPa.setter
    def molskgcatPa(self, value):
        self.si = utils.to_float(value)

    @property
    def molhgcatbar(self):
        return self.si * 3600 / 1000 * (100000**self.order)

    @molhgcatbar.setter
    def molhgcatbar(self, value):
        self.si = utils.to_float(value / 3600 * 1000 / (100000**self.order))

    @property
    def molmingcatbar(self):
        return self.si * 60 / 1000 * (100000**self.order)

    @molmingcatbar.setter
    def molmingcatbar(self, value):
        self.si = utils.to_float(value / 60 * 1000 / (100000**self.order))

    @property
    def molskgcatbar(self):
        return self.si * (100000**self.order)

    @molskgcatbar.setter
    def molskgcatbar(self, value):
        self.si = utils.to_float(value / (100000**self.order))

    @property
    def Lgcatmin(self):
        from . import R

        return self.molmingcatbar * (R.LbarKmol * self.T.K) ** self.order

    @Lgcatmin.setter
    def Lgcatmin(self, value):
        from . import R

        self.molmingcatbar = value / ((R.LbarKmol * self.T.K) ** self.order)

    # def get_Lgcatmin(self, T):
    #     return self.molmingcatbar * (R.LbarKmol * T.K) ** self.order

    # def set_Lgcatmin(self, value, T):
    #     self.molmingcatbar = value / ((R.LbarKmol * T.K) ** self.order)

    def __call__(self, T):
        from . import R

        self.T = T
        if self.Tref is not None:
            si = self.kref_si * np.exp(
                -self.Ea.si / R.si * (1 / T.si - 1 / self.Tref.si)
            )
        else:
            si = self.si * np.exp(-self.Ea.si / R.si / T.si)
        new_rate_constant = RateConstant(
            Ea=self.Ea, order=self.order, Tref=self.Tref, si=si
        )
        new_rate_constant.T = T
        return new_rate_constant

    # def set_temp(self, T):
    #     self.T = T
    #     return self



import numpy as np

from ..kinetic_model import KineticModel
from catrxneng.utils import vant_hoff_eqn
from catrxneng.reactions import RWGS, CO2ToCH3OH, Sabatier
from catrxneng.quantities import *


class Ghosh2021(KineticModel):
    def __init__(self, limiting_reactant="co2", site_model="single", T=None, kref=None, Ea=None):
        self.site_model = site_model
        self.Tref = Temperature(C=300)
        self.components = ["co2", "h2", "ch3oh", "h2o", "co", "ch4", "inert"]
        self.carbon_atoms = np.array([1, 0, 1, 0, 1, 1, 0])
        self.reactions = {
            "co2_to_meoh": CO2ToCH3OH(),
            "rwgs": RWGS(),
            "sabatier": Sabatier(),
        }
        self.kref = kref
        if self.kref is None:
            self.kref = np.array([6.9e-4, 1.8e-3, 1.1e-4])
        self.Ea = Ea
        if self.Ea is None:
            self.Ea = np.array([35.7, 54.5, 42.5])
        self.order = np.array([2.0, 1.5, 1.0])
        self.comp_idx = {comp: i for i, comp in enumerate(self.components)}
        self.rxn_idx = {rxn: i for i, rxn in enumerate(self.reactions)}
        super().__init__(limiting_reactant, T)
        self.stoich_coeff = np.array(
            [
                [-1, -3, 1, 1, 0, 0, 0],
                [-1, -1, 0, 1, 1, 0, 0],
                [-1, -4, 0, 2, 0, 1, 0],
            ]
        )

    def compute_temp_dependent_constants(self):
        self.K_h2 = vant_hoff_eqn(0.76, Energy(kJmol=-12.5), self.T, self.Tref).si
        self.K_co2 = vant_hoff_eqn(0.79, Energy(kJmol=-25.9), self.T, self.Tref).si
        self.K_co2_to_meoh = self.reactions["co2_to_meoh"].Keq(self.T)
        self.K_rwgs = self.reactions["rwgs"].Keq(self.T)
        self.K_sabatier = self.reactions["sabatier"].Keq(self.T)
        self.Keq = np.array([self.K_co2_to_meoh, self.K_rwgs, self.K_sabatier])
        self.k = np.array(
            [
                RateConstant(
                    molskgcatbar=kref, Ea=Energy(kJmol=Ea), Tref=self.Tref, order=order
                )(self.T).molhgcatbar
                for kref, Ea, order in zip(self.kref, self.Ea, self.order)
            ]
        )

    def rate_equations(self, p_array):
        p_co2 = p_array[0]  # co2
        p_h2 = p_array[1]  # h2
        p_ch3oh = p_array[2]  # ch3oh
        p_h2o = p_array[3]  # h2o
        p_co = p_array[4]  # co
        p_ch4 = p_array[5]  # ch4

        inhib = (1 + self.K_co2 * p_co2 + np.sqrt(self.K_h2 * p_h2)) ** 2

        fwd = p_co2 * p_h2**3
        rev = p_ch3oh * p_h2o / self.K_co2_to_meoh
        numerator = self.k[0] * (fwd - rev) / (p_h2**2)
        r1 = numerator / inhib

        fwd = p_co2 * p_h2
        rev = p_co * p_h2o / self.K_rwgs
        numerator = self.k[1] * (fwd - rev) / np.sqrt(p_h2)
        r2 = numerator / inhib

        numerator = p_ch4 * p_h2o**2
        denom = p_co2 * p_h2**4 * self.K_sabatier
        if denom == 0:
            r3 = 0
        else:
            frac = (1 - numerator / denom) / inhib
            r3 = self.k[2] * np.sqrt(p_co2 * p_h2) * frac

        return np.array(
            [
                -r1 - r2 - r3,  # co2
                -3 * r1 - r2 - 4 * r3,  # h2
                r1,  # ch3oh
                r1 + r2 + 2 * r3,  # h2o
                r2,  # co
                r3,  # ch4
                0.0,  # inert
            ]
        )



from .quantity import Quantity
from ..species import GasMixture
from catrxneng.utils import *


class SpaceTime(Quantity):

    def __init__(self, gas_mixture: GasMixture = None, **kwargs):
        self.gas_mixture = gas_mixture
        super().__init__(**kwargs)

    @property
    def molskgcat(self):
        return self.si

    @molskgcat.setter
    def molskgcat(self, value):
        self.si = to_float(value)

    @property
    def molhgcat(self):
        return self.si / 3600 * 1000

    @molhgcat.setter
    def molhgcat(self, value):
        self.si = to_float(value) * 3600 / 1000

    @property
    def smLhgcat(self):
        return self.si / 3600 / 22.4

    @smLhgcat.setter
    def smLhgcat(self, value):
        self.si = to_float(value) * 3600 * 22.4

    @property
    def h(self):
        try:
            return self.si / 1000 * self.gas_mixture.avg_mol_weight * 3600
        except TypeError:
            raise AttributeError("Spacetime has no gas mixture assigned.")

    @h.setter
    def h(self, value):
        try:
            self.si = value * 1000 / self.gas_mixture.avg_mol_weight / 3600
        except TypeError:
            raise AttributeError("Spacetime has no gas mixture assigned.")



from .species import Species
from ..quantities import *


class C3H6(Species):
    def __init__(self, T=None):
        self.mol_weight = 42.081
        self.c_atoms = 3
        self.h_atoms = 6
        super().__init__(T)



from catrxneng import utils
from .quantity import Quantity


class EquilibriumConstant(Quantity):
    def __init__(self, order, **kwargs):
        self.order = order
        super().__init__(**kwargs)

    @property
    def Pa(self):
        return self.si

    @Pa.setter
    def Pa(self, value):
        self.si = utils.to_float(value)

    @property
    def bar(self):
        return self.si * (100000**self.order)

    @bar.setter
    def bar(self, value):
        self.si = utils.to_float(value / (100000**self.order))



from .species import Species

from catrxneng import utils


class H2O(Species):
    def __init__(self, T=None):
        from ..quantities import Energy, Entropy, Temperature

        self.mol_weight = 18
        self.Hf_298_liq = Energy(kJmol=-285.83)
        self.S_298_liq = Entropy(JmolK=69.95)
        self.Hf_298_gas = Energy(kJmol=-241.83)
        self.S_298_gas = Entropy(JmolK=188.84)
        self.boiling_temp = Temperature(C=100)
        self.dH_vap = Energy(kJmol=40.657)
        self.dS_vap = Entropy(JmolK=109)
        # self.Cp_liq = HeatCapacity(JmolK=75.3)
        self.nist_thermo_params = [
            {
                "min_temp_K": 298,
                "max_temp_K": 500,
                "phase": "liq",
                "A": -203.606,
                "B": 1523.29,
                "C": -3196.413,
                "D": 2474.455,
                "E": 3.855326,
                "F": -256.5487,
                "G": -488.7163,
                "H": -285.8304,
            },
            {
                "min_temp_K": 500,
                "max_temp_K": 1700,
                "phase": "gas",
                "A": 30.09200,
                "B": 6.832514,
                "C": 6.793435,
                "D": -2.534480,
                "E": 0.082139,
                "F": -250.8810,
                "G": 223.3967,
                "H": -241.8264,
            },
        ]
        super().__init__(T)

    def Cp_gas(self, T=None):
        from ..quantities import HeatCapacity

        if not T:
            T = self.T
        params = self.nist_thermo_params[1]
        if params["min_temp_K"] <= T.K <= params["max_temp_K"]:
            return utils.Cp_shomate(T, params)
        elif 298 <= T.K <= 500:
            return HeatCapacity(JmolK=34.52)
        else:
            raise ValueError("Invalid temperature for H2O Cp_gas.")

    def Cp_liq(self, T=None):
        if not T:
            T = self.T
        params = self.nist_thermo_params[0]
        if params["min_temp_K"] <= T.K <= params["max_temp_K"]:
            return utils.Cp_shomate(T, params)
        else:
            raise ValueError("Invalid temperature for H2O Cp_liq.")



import numpy as np
from scipy.optimize import minimize

from catrxneng.quantities import *


class Reaction:
    def __init__(self, T, limiting_reactant):
        self.limiting_reactant = limiting_reactant
        self.T = T
        self.fug_coeff = Unitless(
            si=np.ones(len(self.components)), keys=list(self.components.keys())
        )
        # self.update()

    @property
    def active_components(self):
        return {key: self.components[key] for key in self.components if key != "inert"}

    @property
    def stoich_coeff_active(self):
        return Unitless(
            si=self.stoich_coeff.si[self.stoich_coeff.si != 0],
            keys=list(self.active_components.keys()),
        )

    @property
    def dH_rxn_298_gas(self):
        Hf_298_gas = np.array(
            [self.components[key].Hf_298_gas.si for key in self.active_components]
        )
        dHr_298_gas = np.sum(Hf_298_gas * self.stoich_coeff_active.si)
        return Energy(si=dHr_298_gas)

    def dH_rxn_gas(self, T=None):
        if not T:
            T = self.T
        Hf_gas = np.array(
            [self.components[key].Hf_gas(T).si for key in self.active_components]
        )
        dHr_gas = np.sum(Hf_gas * self.stoich_coeff_active.si)
        return Energy(si=dHr_gas)

    @property
    def dS_rxn_298_gas(self):
        S_298_gas = np.array(
            [self.components[key].S_298_gas.si for key in self.active_components]
        )
        dSr_298_gas = np.sum(S_298_gas * self.stoich_coeff_active.si)
        return Entropy(si=dSr_298_gas)

    def dS_rxn_gas(self, T=None):
        if not T:
            T = self.T
        Sf_gas = np.array(
            [self.components[key].Sf_gas(T).si for key in self.active_components]
        )
        dSr_gas = np.sum(Sf_gas * self.stoich_coeff_active.si)
        return Energy(si=dSr_gas)

    @property
    def dG_rxn_298_gas(self):
        T = Temperature(K=298)
        return self.dH_rxn_298_gas - T * self.dS_rxn_298_gas

    def dG_rxn(self, T=None):
        if not T:
            T = self.T
        return self.dH_rxn(T) - T * self.dS_rxn(T)

    def dG_rxn_gas(self, T=None):
        if not T:
            T = self.T
        return self.dH_rxn_gas(T) - T * self.dS_rxn_gas(T)

    def Keq(self, T=None):
        if not T:
            T = self.T
        return np.exp(-self.dG_rxn(T) / (R * T)).si

    def check_components(self, p0, allow_component_mismatch):
        if p0.keys != list(self.components.keys()) and not allow_component_mismatch:
            raise ValueError("Partial pressure keys do not match reaction components.")

    def equilibrate(self, p0, T, allow_component_mismatch=False):
        self.check_components(p0,allow_component_mismatch=allow_component_mismatch)
        if allow_component_mismatch:
            p0_bar = [p0[comp].bar for comp in list(self.components.keys())]
            p0 = Pressure(bar=p0_bar, keys=list(self.components.keys()))
        P = np.sum(p0)
        initial_total_moles = Moles(si=100)
        initial_molfrac = p0 / P
        initial_moles = initial_molfrac * initial_total_moles
        std_state_fugacity = Pressure(atm=np.ones(len(self.components)))

        def objective(extent):
            extent = Moles(si=extent)
            moles = initial_moles + extent * self.stoich_coeff
            total_moles = np.sum(moles)
            molfrac = moles / total_moles
            fugacity = molfrac * self.fug_coeff * P
            activity = fugacity / std_state_fugacity
            Ka = np.prod(activity**self.stoich_coeff)
            # Kx = np.prod(molfrac**self.stoich_coeff)
            # Kphi = np.prod(self.fug_coeff**self.stoich_coeff)
            # Kp = np.prod(P**self.stoich_coeff)
            # Kf0 = np.prod(std_state_fugacity**self.stoich_coeff)
            return ((Ka - self.Keq(T)) ** 2).si * 1e5

        adj_init_mol_reactants = np.array(
            [
                mol / stoich_coeff
                for mol, stoich_coeff in zip(initial_moles.si, self.stoich_coeff.si)
                if stoich_coeff < 0
            ]
        )
        min_extent = 1e-5
        bounds = [(min_extent, np.min(-adj_init_mol_reactants) * (1 - min_extent))]
        initial_guess = -0.5 * initial_moles.si[0] / self.stoich_coeff.si[0]
        result = minimize(
            objective,
            initial_guess,
            bounds=bounds,
            options={"ftol": 1e-10},
        )
        if result.success:
            self.extent = Moles(si=result.x[0])
            moles = initial_moles + self.extent * self.stoich_coeff
            self.eq_conversion = (
                initial_moles[self.limiting_reactant] - moles[self.limiting_reactant]
            ) / initial_moles[self.limiting_reactant]
            total_moles = Moles(si=np.sum(moles.si))
            self.eq_molfrac = moles / total_moles
        else:
            raise ValueError("Optimization failed: " + result.message)



from .quantity import Quantity
from .energy import Energy
from .temperature import Temperature
from .entropy import Entropy
from .pressure import Pressure
from .unitless import Unitless
from .mass import Mass
from .whsv import WHSV
from .inv_whsv import InvWHSV
from .moles import Moles
from .molar_flow_rate import MolarFlowRate
from .reaction_rate import ReactionRate
from .rate_constant import RateConstant
from .vol_flow_rate import VolumetricFlowRate
from .inv_pressure import InversePressure
from .time_delta import TimeDelta
from .heat_capacity import HeatCapacity
from .arbitrary_units import ArbitraryUnits
from .equilibrium_constant import EquilibriumConstant
from .concentration import Concentration
from .spacetime import SpaceTime

from .gas_constant import GasConstant
R = GasConstant()

std_state_fugacity = Pressure(atm=1)


import numpy as np
from scipy.integrate import solve_ivp

from .. import utils
from ..quantities import *
from .reactor import Reactor
from catrxneng.species.gas_mixture import GasMixture


class PFR(Reactor):
    def __init__(
        self,
        kinetic_model,
        p0,
        whsv=None,
        T=None,
        limiting_reactant=None,
        mcat=None,
        F0=None,
    ):
        self.kinetic_model = kinetic_model
        self.p0 = p0
        self.P = np.sum(p0)
        self.P_bar = self.P.bar
        self.y0 = utils.divide(p0, self.P)
        self.whsv = whsv
        self.T = T
        if self.T is None:
            self.T = kinetic_model.T
        self.limiting_reactant = limiting_reactant
        if not self.limiting_reactant:
            self.limiting_reactant = self.kinetic_model.limiting_reactant
        self.mcat = mcat
        if not mcat:
            self.mcat = Mass(g=1)
        if self.whsv:
            if self.whsv.gas_mixture is None:
                self.whsv.gas_mixture = GasMixture(p=p0)
            self.Ft0_active = self.whsv * self.mcat
            self.Ft0 = self.Ft0_active / (1 - self.y0["inert"])
            self.F0 = self.y0 * self.Ft0
        else:
            self.F0 = F0
            self.Ft0 = MolarFlowRate(molh=np.sum(self.F0.molh))
            self.Ft0_active = self.Ft0 - self.F0["inert"]
            # self.Ft0_active = MolarFlowRate(
            #     molh=np.sum(
            #         [self.F0[comp].molh for comp in self.p0.keys if comp != "inert"]
            #     )
            # )
            gas_mixture = GasMixture(p=self.p0)
            self.whsv = WHSV(
                molhgcat=self.Ft0_active.molh / self.mcat.g, gas_mixture=gas_mixture
            )

        self.check_components()

    def dFdw(self, w, F):
        p_array = F / np.sum(F) * self.P_bar
        return self.kinetic_model.rate_equations(p_array)

    def dfdx(self, x, f):
        # Ft = np.sum(self.F0[self.limiting_reactant].si * f)
        Ft = np.sum(self.Ft0_active.si * f)
        p = self.P.si * self.Ft0_active.si * f / Ft
        p = Pressure(si=p, keys=self.p0.keys.copy())
        return (
            self.mcat.si
            / self.Ft0_active.si
            * np.array(
                [
                    rate(p, self.T).si
                    for rate in self.kinetic_model.rate_equations.values()
                ]
            )
        )

    def solve_dimensional(self, points, method):
        w_span = (0, self.mcat.g)
        w_eval = np.linspace(0, self.mcat.g, points)
        # F0 = getattr(self.F0, self.units["molar_flow_rate"])
        F0 = self.F0.molh
        solution = solve_ivp(self.dFdw, w_span, F0, t_eval=w_eval, method=method)
        self.w = Mass(g=solution.t)
        # kwargs = {self.units["molar_flow_rate"]: solution.y}
        # kwargs["keys"] = self.kinetic_model.components
        self.F = MolarFlowRate(molh=solution.y, keys=self.F0.keys)

    def solve_dimensionless(self, points=1000, method="RK45", rtol=1e-3, atol=1e-6):
        x_span = (0, 1)
        x_eval = np.linspace(x_span[0], x_span[1], points)
        f0 = self.F0.si / self.Ft0_active.si
        solution = solve_ivp(
            self.dfdx, x_span, f0, t_eval=x_eval, method=method, rtol=rtol, atol=atol
        )
        self.x = solution.t
        self.f = solution.y
        self.w = Mass(si=(self.x * self.mcat.si))
        F = self.Ft0_active.si * self.f
        self.F = MolarFlowRate(si=F, keys=self.F0.keys)

    def solve(self, points=100, nondimensionalize=False, method="LSODA"):
        if nondimensionalize:
            self.solve_dimensionless(points=points, method=method)
        else:
            self.solve_dimensional(points=points, method=method)
        self.y = utils.divide(self.F, np.sum(self.F, axis=0))
        self.conversion = utils.divide(
            self.F0[self.limiting_reactant] - self.F[self.limiting_reactant],
            self.F0[self.limiting_reactant],
        )
        # self.yield_ = self.F / self.F0[self.limiting_reactant]
        spacetime = utils.divide(self.w.g, self.Ft0.smLh)
        self.spacetime = SpaceTime(smLhgcat=spacetime)
        self.whsv_array = WHSV(smLhgcat=utils.divide(1, spacetime))
        vol_flow_rate = self.Ft0.si * R.si * self.T.si / self.P.si
        self.vol_flow_rate = VolumetricFlowRate(si=vol_flow_rate)



from ..quantities import *


class Species:

    def __init__(self, T=None):
        self.T = T
        # self.update()

    def _get_thermo_params(self, T):
        for thermo_params in self.nist_thermo_params:
            if thermo_params["min_temp_K"] <= T.K <= thermo_params["max_temp_K"]:
                return thermo_params
            raise ValueError(
                f"Temperature outside range for {type(self).__name__} thermodynamic parameters."
            )



from catrxneng import utils
from .species import Species
from ..quantities import *


class CO2(Species):
    def __init__(self, T=None):
        self.mol_weight = 44
        self.Hf_298_gas = Energy(kJmol=-393.51)
        self.S_298_gas = Entropy(JmolK=213.79)
        self.nist_thermo_params = [
            {
                "min_temp_K": 298,
                "max_temp_K": 1200,
                "phase": "gas",
                "A": 24.99735,
                "B": 55.18696,
                "C": -33.69137,
                "D": 7.948387,
                "E": -0.136638,
                "F": -403.6075,
                "G": 228.2431,
                "H": -393.5224,
            }
        ]
        super().__init__(T)

    def Hf_gas(self, T=None):
        if not T:
            T = self.T
        return utils.Hf_shomate(T, self.nist_thermo_params[0])

    def S_gas(self, T=None):
        if not T:
            T = self.T
        return utils.S_shomate(T, self.nist_thermo_params[0])

    def Cp_gas(self, T=None):
        if not T:
            T = self.T
        return utils.Cp_shomate(T, self.nist_thermo_params[0])
    


from .reaction import Reaction
from .. import species
from ..quantities import *
from .rwgs import RWGS


class WGS(Reaction):
    def __init__(self, T=None, limiting_reactant="co"):
        self.components = {
            "co": species.CO(T=T),
            "h2o": species.H2O(T=T),
            "co2": species.CO2(T=T),
            "h2": species.H2(T=T),
            "inert": species.Ar(T=T),
        }
        self.stoich_coeff = Unitless(
            si=[-1.0, -1.0, 1.0, 1.0, 0.0], keys=list(self.components.keys())
        )
        super().__init__(T=Tk, limiting_reactant=limiting_reactant)

    def Keq(self, T=None):
        if not T:
            T = self.T
        return 1 / RWGS().Keq(T)



import plotly.graph_objects as go, numpy as np
from catrxneng.utils import plot_info_box


class BasicPlot:

    def __init__(self, title=None):
        self.fig = go.Figure()
        self.title = title
        self.right_axis = False

    def add_trace(self, x, y, name=None, mode="lines", yaxis="y1", hover_label=None, line_dash=None):
        if isinstance(y, (float, int)):
            if x is not None:
                xmin = np.min(x)
                xmax = np.max(x)
            x = [xmin, xmax]
            y = [y, y]
        hovertemplate = None
        if hover_label:
            hovertemplate = "<b>%{text}</b><br>X: %{x}<br>Y: %{y}<extra></extra>"
        trace_kwargs = dict(
            x=x,
            y=y,
            mode=mode,
            yaxis=yaxis,
            name=name,
            text=hover_label,
            hovertemplate=hovertemplate,
        )
        if line_dash:
            trace_kwargs["line"] = dict(dash=line_dash)
        trace = go.Scatter(**trace_kwargs)
        self.fig.add_trace(trace)

    def add_vertical_line(self, x, line_dash="dash", line_color="black", line_width=2):
        """
        Adds a vertical dashed line at the specified x-value.

        Args:
            x_value (float): The x-coordinate where the vertical line will be added.
            line_dash (str): The dash style of the line (default is "dash").
            line_color (str): The color of the line (default is "black").
            line_width (int): The width of the line (default is 2).
        """
        self.fig.add_shape(
            type="line",
            x0=x,
            x1=x,
            y0=0,
            y1=1,
            xref="x",
            yref="paper",
            line=dict(
                dash=line_dash,
                color=line_color,
                width=line_width,
            ),
        )

    def render(
        self,
        xlabel,
        ylabel,
        xrange=None,
        yrange=None,
        ylabel2=None,
        yrange2=None,
        info_text=None,
    ):
        if xrange is None:
            xrange = [None, None]
        if yrange is None:
            yrange = [None, None]
        if yrange2 is None:
            yrange2 = [None, None]

        top = 50
        width = 700
        if info_text is None:
            bottom = 50
            height = 400
        else:
            bottom = 110
            height = None

        yaxis2 = None
        if ylabel2:
            yaxis2 = dict(
                title=f"<b>{ylabel2}</b>",
                range=yrange2,
                showline=True,
                linecolor="black",
                linewidth=2,
                mirror=True,
                nticks=9,
                overlaying="y",
                side="right",
            )
        self.fig.update_layout(
            title=dict(text=f"<b>{self.title}</b>", x=0.5),
            xaxis_title=f"<b>{xlabel}</b>",
            yaxis_title=f"<b>{ylabel}</b>",
            width=width,
            height=height,
            margin=dict(t=top, b=bottom),
            yaxis=dict(
                range=yrange,
                showline=True,
                linecolor="black",
                linewidth=2,
                mirror=True,
                nticks=9,
            ),
            yaxis2=yaxis2,
            xaxis=dict(
                range=xrange,
                showline=True,
                linecolor="black",
                linewidth=2,
                mirror=True,
            ),
            legend=dict(x=1.05, y=1, xanchor="left", yanchor="top"),
            annotations=plot_info_box(info_text),
            plot_bgcolor="white",
            paper_bgcolor="white",
        )
        self.fig.show()



import numpy as np

from ..kinetic_model import KineticModel
from catrxneng.utils import vant_hoff_eqn
from catrxneng.quantities import *


class Ghosh2022Mod(KineticModel):
    def __init__(self, limiting_reactant="ch3oh", T=None, kref=None, Ea=None):
        self.Tref = Temperature(C=320)
        self.components = [
            "ch3oh",
            "h2",
            "h2o",
            "c2h4",
            "c3h6",
            "c4h8",
            "c5h10",
            "c2h6",
            "c3h8",
            "c4h10",
            "c5h12",
            "inert",
        ]
        self.carbon_atoms = np.array([1, 0, 0, 2, 3, 4, 5, 2, 3, 4, 5, 0])
        self.kref = kref
        if self.kref is None:
            k2 = 0.001
            self.kref = np.array([1000, 800, 600, 400, k2, k2, k2, k2])
        self.Ea = Ea
        if self.Ea is None:
            Ea1 = 70.0
            Ea2 = 110.0
            self.Ea = np.array([Ea1, Ea1, Ea1, Ea1, Ea2, Ea2, Ea2, Ea2])
        self.order = np.array([1, 1, 1, 1, 2, 2, 2, 2])
        self.comp_idx = {comp: i for i, comp in enumerate(self.components)}
        super().__init__(limiting_reactant, T)

    def compute_temp_dependent_constants(self):
        self.K_ch3oh = vant_hoff_eqn(1.3e1, Energy(kJmol=-2e-1), self.T, self.Tref).si
        self.K_h2o = vant_hoff_eqn(1e1, Energy(kJmol=-2e-1), self.T, self.Tref).si
        self.k = np.array(
            [
                RateConstant(
                    molskgcatbar=kref, Ea=Energy(kJmol=Ea), Tref=self.Tref, order=order
                )(self.T).molhgcatbar
                for kref, Ea, order in zip(self.kref, self.Ea, self.order)
            ]
        )

    def rate_equations(self, p_array):
        p_ch3oh = p_array[self.comp_idx["ch3oh"]]
        p_h2 = p_array[self.comp_idx["h2"]]
        p_h2o = p_array[self.comp_idx["h2o"]]
        p_c2h4 = p_array[self.comp_idx["c2h4"]]
        p_c3h6 = p_array[self.comp_idx["c3h6"]]
        p_c4h8 = p_array[self.comp_idx["c4h8"]]
        p_c5h10 = p_array[self.comp_idx["c5h10"]]
        inhib = 1 + self.K_ch3oh * p_ch3oh + self.K_h2o * p_h2o

        r_c2h4 = self.k[0] * p_ch3oh / inhib
        r_c3h6 = self.k[1] * p_ch3oh / inhib
        r_c4h8 = self.k[2] * p_ch3oh / inhib
        r_c5h10 = self.k[3] * p_ch3oh / inhib
        r_c2h6 = self.k[4] * p_c2h4 * p_h2 / inhib
        r_c3h8 = self.k[5] * p_c3h6 * p_h2 / inhib
        r_c4h10 = self.k[6] * p_c4h8 * p_h2 / inhib
        r_c5h12 = self.k[7] * p_c5h10 * p_h2 / inhib

        return np.array(
            [
                -2 * r_c2h4 - 3 * r_c3h6 - 4 * r_c4h8 - 5 * r_c5h10,  # meoh
                -r_c2h4 - r_c3h6 - r_c4h8 - r_c5h10,  # h2
                2 * r_c2h4 + 3 * r_c3h6 + 4 * r_c4h8 + 5 * r_c5h10,  # h2o 
                r_c2h4 - r_c2h6, # c2h4
                r_c3h6 - r_c3h8, # c3h6
                r_c4h8 - r_c4h10, # c4h8
                r_c5h10 - r_c5h12, # c5h10
                r_c2h6, # c2h6
                r_c3h8, # c3h8
                r_c4h10, # c4h10
                r_c5h12, # c5h12
                0.0,  # inert
            ]
        )



from .species import Species
from ..quantities import *


class C2H4(Species):
    def __init__(self, T=None):
        self.mol_weight = 28
        self.c_atoms = 2
        self.h_atoms = 4
        self.Hf_298 = Energy(kJmol=52.4)
        self.nist_thermo_params = [{
            "min_temp_K": 298,
            "max_temp_K": 1200,
            "phase": "gas",
            "A": -6.387880,
            "B": 184.4019,
            "C": -112.9718,
            "D": 28.49593,
            "E": 0.315540,
            "F": 48.17332,
            "G": 163.1568,
            "H": 52.46694
        }]
        super().__init__(T)



from .quantity import Quantity
from catrxneng.utils import *


class Concentration(Quantity):

    @property
    def molm3(self):
        return self.si

    @molm3.setter
    def molm3(self, value):
        self.si = to_float(value)

    @property
    def molL(self):
        return self.si / 1000

    @molL.setter
    def molL(self, value):
        self.si = to_float(value * 1000)



from .species import Species


class DME(Species):
    def __init__(self, T=None):
        from ..quantities import Energy, Entropy, Temperature

        self.mol_weight = 46.069
        self.c_atoms = 2
        self.h_atoms = 6
        self.o_atoms = 1
        # self.Hf_298_liq = Energy(kJmol=-238.4)
        # self.S_298_liq = Entropy(JmolK=127.19)
        # self.Hf_298_gas = Energy(kJmol=-205)
        # self.dH_vap = Energy(kJmol=35.21)
        # self.dS_vap = Entropy(JmolK=104.6)
        # self.boiling_temp = Temperature(C=64.7)
        super().__init__(T)

    # def Cp_liq(self, T=None):
    #     from ..quantities import HeatCapacity

    #     if not T:
    #         T = self.T
    #     cp = 0.1955 * T.K + 22.419
    #     return HeatCapacity(JmolK=cp)

    # def Cp_gas(self, T=None):
    #     from ..quantities import HeatCapacity

    #     if not T:
    #         T = self.T
    #     cp = 0.0656 * T.K + 26.21
    #     return HeatCapacity(JmolK=cp)



import scipy.integrate as integrate

from .reaction import Reaction
from ..quantities import Unitless, Temperature, Energy, Entropy
from ..species import CO2, H2, CH4, H2O, Ar


class Sabatier(Reaction):
    def __init__(self, T=None, limiting_reactant="co2"):
        self.T = T
        self.components = {
            "co2": CO2(T=T),
            "h2": H2(T=T),
            "ch4": CH4(T=T),
            "h2o": H2O(T=T),
            "inert": Ar(T=T),
        }
        self.stoich_coeff = Unitless(
            si=[-1.0, -4.0, 1.0, 2.0, 0.0], keys=list(self.components.keys())
        )
        super().__init__(T=T, limiting_reactant=limiting_reactant)

    def dCp_1(self, T_K):
        T = Temperature(K=T_K)
        dCp = (
            CH4().Cp_gas(T) + 2 * H2O().Cp_liq(T) - CO2().Cp_gas(T) - 4 * H2().Cp_gas(T)
        )
        return dCp.JmolK

    def dCp_2(self, T_K):
        T = Temperature(K=T_K)
        dCp = (
            CH4().Cp_gas(T) + 2 * H2O().Cp_gas(T) - CO2().Cp_gas(T) - 4 * H2().Cp_gas(T)
        )
        return dCp.JmolK

    @property
    def dH_rxn_298(self):
        return (
            CH4().Hf_298_gas
            + 2 * H2O().Hf_298_liq
            - CO2().Hf_298_gas
            - 4 * H2().Hf_298_gas
        )

    @property
    def dH_rxn_298_gas(self):
        return (
            CH4().Hf_298_gas
            + 2 * H2O().Hf_298_gas
            - CO2().Hf_298_gas
            - 4 * H2().Hf_298_gas
        )

    @property
    def dS_rxn_298(self):
        return (
            CH4().S_298_gas + 2 * H2O().S_298_liq - CO2().S_298_gas - 4 * H2().S_298_gas
        )

    def dH_rxn(self, T=None):
        Tb_h2o = H2O().boiling_temp.K
        if T.K < Tb_h2o:
            dHr = self.dH_rxn_298.Jmol
            dHr += integrate.quad(self.dCp_1, 298, T.K)[0]
            return Energy(Jmol=dHr)
        if T.K >= Tb_h2o:
            dHr = self.dH_rxn_298.Jmol
            dHr += integrate.quad(self.dCp_1, 298, Tb_h2o)[0]
            dHr += H2O().dH_vap.Jmol
            dHr += integrate.quad(self.dCp_2, Tb_h2o, T.K)[0]
            return Energy(Jmol=dHr)

    def dS_rxn(self, T=None):
        Tb_h2o = H2O().boiling_temp.K
        integrand1 = lambda T_K: self.dCp_1(T_K) / T_K
        integrand2 = lambda T_K: self.dCp_2(T_K) / T_K
        if T.K < Tb_h2o:
            dSr = self.dS_rxn_298.JmolK
            dSr += integrate.quad(integrand1, 298, T.K)[0]
            return Entropy(JmolK=dSr)
        if T.K >= Tb_h2o:
            dSr = self.dS_rxn_298.JmolK
            dSr += integrate.quad(integrand1, 298, Tb_h2o)[0]
            dSr += H2O().dS_vap.JmolK
            dSr += integrate.quad(integrand2, Tb_h2o, T.K)[0]
            return Entropy(JmolK=dSr)
